#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <assert.h>
#include <iostream>
#include <string.h>

// Project includes
#include "RtAudio.h"
#include "chuck_fft.h"
#include "Common.h"


/*********************************************************************************
  Operating System Abstraction
 **********************************************************************************/
#ifdef __MACOSX_CORE__
// note: for mac only
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#ifdef WIN32
#include <glut.h>
#else
#include <GL/glut.h>
#endif 
#endif

// Operating system dependent thread functionality.
#if defined(__WINDOWS_DS__) || defined(__WINDOWS_ASIO__)
#include <windows.h>
#include <process.h>

// Mutex related definition
typedef CRITICAL_SECTION OsMutex;
#define MUTEX_INITIALIZE(A) InitializeCriticalSection(A)
#define MUTEX_DESTROY(A)    DeleteCriticalSection(A)
#define MUTEX_LOCK(A)       EnterCriticalSection(A)
#define MUTEX_UNLOCK(A)     LeaveCriticalSection(A)

#define OS_SLEEP( X ) Sleep(X)
#define RT_AUDIO_API_TYPE     (RtAudio::WINDOWS_DS)

#endif

#if defined(__LINUX_ALSA__) || defined(__UNIX_JACK__) || defined(__LINUX_OSS__) || defined(__MACOSX_CORE__)
#include <unistd.h>  
#include <pthread.h>
// Mutex related definitions
typedef pthread_mutex_t OsMutex;
#define MUTEX_INITIALIZE(A) pthread_mutex_init(A, NULL)
#define MUTEX_DESTROY(A)    pthread_mutex_destroy(A)
#define MUTEX_LOCK(A)       pthread_mutex_lock(A)
#define MUTEX_UNLOCK(A)     pthread_mutex_unlock(A)

#define OS_SLEEP( X ) usleep(X)

#if defined (__LINUX_ALSA__)
#define RT_AUDIO_API_TYPE     (RtAudio::LINUX_ALSA)
#elif defined (__UNIX_JACK__)
#define RT_AUDIO_API_TYPE     (RtAudio::UNIX_JACK)
#else
#error
#endif

#endif

/*********************************************************************************
  Macros
 **********************************************************************************/

// Define some colors
#define GET_COLOR3f(r,g,b,fact)         (float)((fact)*(r))/255.0, \
  (float)((fact)*(g))/255.0,\
(float)((fact)*(b))/255.0
#define COLOR_DECAY_RATE                0.97
#define PURPLE(fact)                    GET_COLOR3f(171,130,255,fact)
#define YELLOW(fact)                    GET_COLOR3f(255,255,0,fact)
#define ORANGE(fact)                    GET_COLOR3f(255,127,0,fact)
#define GREEN(fact)                     GET_COLOR3f(202,255,112,fact)
#define LIGHT_GREEN(fact)               GET_COLOR3f(124,252,0,fact)
#define PINK(fact)                      GET_COLOR3f(255,20,147,fact)

#define WATERFALL_COLOR                 ORANGE
#define WINDOWED_COLOR                  PINK
#define AUTOCORR_COLOR                  PINK            


#define SAMPLE float
#define MY_PI 3.14159265358979

#ifdef WIN32
#define SAMPLE_RATE     (44100)
#define BUFFER_FRAMES   (1024)
#else
#define SAMPLE_RATE     (48000)
#define BUFFER_FRAMES   (512)
#endif

#define OUT_CHANNELS    (1)
#define IN_CHANNELS     (1)

#define SCALING_FACTOR  (500.0f)
#define FFT_HISTORY_LEN (70)
#define X_START         (-2.0f)
// scaling of the output's amplitude 
#define AMPLITUDE       (1.0)

// Ignore detected pitches if the amplitude is below this
#define PITCH_MAG_THRESH (0.000005)

#define MIN_PITCH        (80)
/*********************************************************************************
  Globals
 *********************************************************************************/
using namespace std;        

// Width and height of the window
GLsizei g_width = 1200;
GLsizei g_height = 768;

float g_scaling_factor = SCALING_FACTOR;

// Global buffers for audio samples
SAMPLE * g_buffer = NULL;
SAMPLE * g_WindowedBuffer = NULL;
SAMPLE * g_HanningBuffer = NULL;
SAMPLE * g_tempBuffer1 = NULL;
SAMPLE * g_tempBuffer2 = NULL;
SAMPLE * g_tempBuffer3 = NULL;
SAMPLE * g_tempBuffer4 = NULL;
SAMPLE * g_zeroPadBuf = NULL;

// zero pad factor
const float zpf = 8;

// Size of the above buffers in samples
long g_bufferSize;

int g_sampleRate = 0;

// Buffers containing FFT history
SAMPLE * g_fftBuffer[FFT_HISTORY_LEN];

// Index into the FFT buffer array
int g_BufIndex = 0;


// Global RtAudio object
RtAudio * pAudio = NULL;

// Mutex for synchronizing between audio and graphics threads
OsMutex g_audioMutex;

// true if the audio stream has been opened
bool streamOpen = false;

// If true to freeze display
bool g_freeze = false;

// If true, frequencies are displayed on a logarithmic scale
// in the waterfall
bool g_logFreq = false;

// If true, the program will output a sine wave with a frequency corresponding
// to the fundemental pitch detected in the input
bool g_outputPitch = false;

float g_pitch = 0.0;

#ifdef WITH_DEBUG
ofstream OUTFILE;
#endif




/*********************************************************************************
  Function Implementations
 **********************************************************************************/

////////////////////////////////////////////////////////////////
// Function Name: AudioCallback
// Description: 
// Parameters: 
// Returns:
////////////////////////////////////////////////////////////////
int AudioCallback( void *outputBuffer, void *inputBuffer,
    unsigned int nFrames,
    double streamTime,
    RtAudioStreamStatus status,
    void *userData )
{
  SAMPLE * pOutputSample = (SAMPLE *) outputBuffer;
  MyFloat32 amplitude = 0.0;
  static float prevPitch = 0.0;

  memset(outputBuffer, 0, sizeof(SAMPLE)*nFrames);

  if (status == 0) {
    SAMPLE * outBuf = (SAMPLE *)outputBuffer;

    MUTEX_LOCK(&g_audioMutex);

    // Make a copy of the input buffer.
    // Note: this assumes single-channel input
    memcpy( g_buffer, (SAMPLE*)inputBuffer, sizeof(SAMPLE)*nFrames );

    MUTEX_UNLOCK(&g_audioMutex);

    if (g_pitch) {
      amplitude = AMPLITUDE;
    } else {
      amplitude *= 0.95;
    }
    if (g_outputPitch && (pOutputSample!= NULL)) {
      MyFloat32 secondsPerSample = 1.0 / g_sampleRate;

      for (unsigned int i=0; i<nFrames; i++) {    
        *pOutputSample  =  sin(streamTime * 2.0 * MY_PI * g_pitch );

        // Advance output sample pointer
        pOutputSample++;        
        // Advance time
        streamTime += secondsPerSample;
      }
    }
  }
  else {
    cerr << "Overflow/Underflow occured, status is " << status << endl;
  }

  return 0;

}


//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void idleFunc( );
void displayFunc( );
void reshapeFunc( GLsizei width, GLsizei height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );

////////////////////////////////////////////////////////////////
// Function Name: cleanUp
// Description: Tries to close the audio stream and release the RtAudio object
//              
// Parameters: none
// Returns: void
////////////////////////////////////////////////////////////////
void cleanUp (void)
{
  DEBUG_PRINTF("cleanup: in\n");
  try {    
    if (pAudio) {// cleanup RtAudio
      pAudio->closeStream();    
      delete pAudio;
      for (int i=0; i<FFT_HISTORY_LEN; i++) {
        delete g_fftBuffer[i];
      }
      delete g_WindowedBuffer;
      delete g_HanningBuffer;
      delete g_tempBuffer1;
      delete g_tempBuffer2;
      delete g_tempBuffer3;
      delete g_tempBuffer4;

      pAudio = NULL;
    }
  } catch (...) {
    cerr << "Exception Caught!" << endl;
    //err.printMessage();
  }
  DEBUG_PRINTF("cleanup: out\n");
}

////////////////////////////////////////////////////////////////
// Function Name: signalHandler
// Description: Cleans up and exits
// Parameters:
// Returns:
////////////////////////////////////////////////////////////////
void signalHandler(int sig)
{
  cleanUp();
  exit(1);
}

void printWelcome()
{
  cout << "\nWelcome to Sound Gaze.\n\n"
          "These keys are available:\n\n"
          "Space\t\tFreeze/Unfreeze the image\n"
          "+/-\t\tMagnify/shrink the waterfall display\n"
          "l\t\tEnable/disable logarithmic frequency mapping\n"
          "p\t\tEnable/disable audio output of detected pitch\n"
          "q\t\tQuit\n\n"
          "Happy Gazing!" << endl;
}


// entry point
int main( int argc, char ** argv )
{       
  unsigned int bufferFrames = BUFFER_FRAMES;
  // initialize GLUT
  glutInit( &argc, argv );
  // double buffer, use rgb color, enable depth buffer
  glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
  // initialize the window size
  glutInitWindowSize( g_width, g_height );
  // set the window postion
  glutInitWindowPosition( 100, 100 );
  // create the window
  glutCreateWindow( "Sound Gaze" );

  // set the idle function - called when idle
  glutIdleFunc( idleFunc );
  // set the display function - called when redrawing
  glutDisplayFunc( displayFunc );
  // set the reshape function - called when client area changes
  glutReshapeFunc( reshapeFunc );
  // set the keyboard function - called on keyboard events
  glutKeyboardFunc( keyboardFunc );
  // set the mouse function - called on mouse stuff
  glutMouseFunc( mouseFunc );  

  glEnable( GL_DEPTH_TEST );


  // Register some signal handlers to allow clean exiting.
  (void) signal(SIGINT, signalHandler);
  (void) signal(SIGABRT, signalHandler);
  (void) signal(SIGTERM, signalHandler);

  // Initialize the mutex
  MUTEX_INITIALIZE(&g_audioMutex);

  /* Initialize audio */
  try {
    RtAudio::StreamParameters outputParameters, inputParameters;


    pAudio = new RtAudio( RT_AUDIO_API_TYPE );

    /* ---- Initialize input and output parameters ------------------- */
    inputParameters.firstChannel = 0;
    inputParameters.nChannels = IN_CHANNELS; 
    outputParameters.firstChannel = 0;
    outputParameters.nChannels = OUT_CHANNELS;
#if defined(__WINDOWS_DS__) || defined(__WINDOWS_ASIO__)        
    outputParameters.deviceId = 0;
    inputParameters.deviceId = 0;
#else // Linux or Mac
    outputParameters.deviceId = pAudio->getDefaultOutputDevice();
    inputParameters.deviceId = pAudio->getDefaultInputDevice();
#endif
    /* ---------------------------------------------------------------- */

    // Open the audio stream
    pAudio->openStream( &outputParameters,
        &inputParameters,
        RTAUDIO_FLOAT32,
        SAMPLE_RATE,
        &bufferFrames, 
        AudioCallback,
        NULL);
    
    g_sampleRate = pAudio->getStreamSampleRate();

    // allocate audio buffers
    g_buffer = new SAMPLE[bufferFrames];
    memset(g_buffer, 0, bufferFrames*sizeof(SAMPLE));
    g_WindowedBuffer = new SAMPLE[bufferFrames];
    memset(g_WindowedBuffer, 0, bufferFrames*sizeof(SAMPLE));
    g_HanningBuffer = new SAMPLE[bufferFrames];
    hanning(g_HanningBuffer, bufferFrames);
    g_tempBuffer1 = new SAMPLE[bufferFrames];
    memset(g_tempBuffer1, 0, bufferFrames*sizeof(SAMPLE));
    g_tempBuffer2 = new SAMPLE[bufferFrames];
    memset(g_tempBuffer2, 0, bufferFrames*sizeof(SAMPLE));
    g_tempBuffer3 = new SAMPLE[bufferFrames];
    memset(g_tempBuffer3, 0, bufferFrames*sizeof(SAMPLE));
    g_tempBuffer4 = new SAMPLE[bufferFrames];
    memset(g_tempBuffer4, 0, bufferFrames*sizeof(SAMPLE));
    //g_zeroPadBuf = new SAMPLE[zpf*bufferFrames];
    //memset(g_zeroPadBuf, 0, zpf*bufferFrames*sizeof(SAMPLE));




    g_bufferSize = bufferFrames;

    // allocate buffers for fft history
    for (int i=0; i<FFT_HISTORY_LEN; i++) {
      g_fftBuffer[i] = new SAMPLE[bufferFrames];
      memset(g_fftBuffer[i] , 0, bufferFrames*sizeof(SAMPLE));
    }

    // Start the stream
    pAudio->startStream();
  } catch (RtError & err) {
    err.printMessage();
    goto cleanup;
  }

  printWelcome();

  // let GLUT handle the current thread from here
  glutMainLoop();

  // if we get here, then stop!
  try {
    pAudio->stopStream();
  } catch( RtError & err ) {
    // do stuff
    err.printMessage();
  }

cleanup:
  pAudio->stopStream();
  delete pAudio;
  delete g_buffer;

  return 0;
}


//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( GLsizei w, GLsizei h )
{
  // save the new window size
  g_width = w; g_height = h;
  // map the view port to the client area
  glViewport( 0, 0, w, h );
  // set the matrix mode to project
  glMatrixMode( GL_PROJECTION );
  // load the identity matrix
  glLoadIdentity( );
  // create the viewing frustum
  gluPerspective(45.0, (GLfloat) w / (GLfloat) h, 1.0, 3000.0 );
  // set the matrix mode to modelview
  glMatrixMode( GL_MODELVIEW );
  // load the identity matrix
  glLoadIdentity( );
  // position the view point
  gluLookAt( 0.0f, 3.0f, 4.0f, 0.0f, 0.0f, -3.0f, 0.0f, 1.0f, 0.0f );
}



//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
  switch( key )
  {
    case ' ':
      // Toggle display freeze
      g_freeze = !g_freeze;
      break;
    case '+':
      // Toggle display freeze
      g_scaling_factor += 50;
      break;
    case '-':
      // Toggle display freeze
      g_scaling_factor -= 50;
      break;

    case 'l':
    case 'L':
      // Toggle display freeze
      g_logFreq = !g_logFreq;
      break;

    case 'p':
    case 'P':
      // Toggle detected pitch audio output
      g_outputPitch = !g_outputPitch;
      break;

    case 'Q':
    case 'q':
      exit(1);
      break;
  }

  glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
  if( button == GLUT_LEFT_BUTTON )
  {
    // when left mouse button is down
    if( state == GLUT_DOWN )
    {
    }
    else
    {
    }
  }
  else if ( button == GLUT_RIGHT_BUTTON )
  {
    // when right mouse button down
    if( state == GLUT_DOWN )
    {
    }
    else
    {
    }
  }
  else
  {
  }

  glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
  // render the scene
  glutPostRedisplay( );
}

//-----------------------------------------------------------------------------
// Name: multBuffers( )
// Desc: multiply bufIn by bufMult and put the result in bufIn
//-----------------------------------------------------------------------------
void multBuffers(SAMPLE* bufIn, SAMPLE* bufMult, long length)
{
  for(int i =0; i<length; i++) {
    bufIn[i] = bufIn[i]*bufMult[i];
  }
}

float getXFraction(int n, float logMax) 
{
  return log10((float)n+10.0)/logMax;
}


float getZeroCrossingRate(SAMPLE* buffer, int numSamples, int sampleRate)
{
  float rate;
  float crossCount=0;
  for(int i=0; i<numSamples-1; i++) {
    if(buffer[i]*buffer[i+1] < 0) {
      crossCount++;
    }
  }

  rate = 0.5*crossCount*sampleRate/numSamples;

  return rate;
}

inline complex complexMultiply(complex * a, complex*b)
{
  complex result;
  result.re = a->re * b->re - a->im * b->im;
  result.im = a->re * b->im + a->im * b->re;
  return result;
}

float computeAutoCorrelation(complex* inFftBuffer, float* outBuffer, int length)
{

  // make copies
  memcpy( g_tempBuffer1, inFftBuffer, sizeof(complex)*length );
  memcpy( g_tempBuffer2, inFftBuffer, sizeof(complex)*length );

  complex* pFirstCopy = (complex*)g_tempBuffer1;
  complex* pSecondCopy = (complex*)g_tempBuffer2;
  complex* pOutComplex = (complex*)outBuffer;

  for(int i=0; i<length; i++){
    // conjugate first copy
    pFirstCopy[i].im *= -1;

    // complex multiply
    pOutComplex[i] = complexMultiply(&pFirstCopy[i], &pSecondCopy[i]);
  }

  return 0.0;
}

void zeroPad(float* pBuffer, int baseLength, int factor)
{
  for(int i = 0; i<(factor-1)*baseLength; i++) {
    pBuffer[i+baseLength] = 0.0;
  }
}

static inline int findMaxIndex(float* pBuffer, int n)
{
  float max = pBuffer[0];
  int maxBin = 0;
  for(int i=0; i<n; i++)
  {
    if(pBuffer[i] > max) {
      max = pBuffer[i];
      maxBin = i;
    }
  }
  return maxBin;
}

float interpolateSpectrum(float a, float b, float c, float binNumber)
{
  float peakLocation = 0.5*(a-c)/(2*a-2*b+c);
  return binNumber+peakLocation;
}

//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
  if (g_freeze) {
    return;
  }

  // Point to where we will place the current spectrum
  SAMPLE* pFFTBuf = g_fftBuffer[g_BufIndex];


  MUTEX_LOCK(&g_audioMutex);
  // ---- Start critical section -----------------------

  memcpy(g_WindowedBuffer, g_buffer, g_bufferSize*sizeof(SAMPLE));

  // ---- End critical section -----------------------
  MUTEX_UNLOCK(&g_audioMutex);

  // Calculate zero crossing
  float zeroCrossPitch = getZeroCrossingRate(g_WindowedBuffer, g_bufferSize, g_sampleRate);

  // Window the signal
  multBuffers(g_WindowedBuffer, g_HanningBuffer, g_bufferSize);

  // Copy windowed signal into current FFT buffer
  memcpy(pFFTBuf, g_WindowedBuffer, g_bufferSize*sizeof(SAMPLE));

  // Do FFT
  rfft(pFFTBuf, g_bufferSize/2, FFT_FORWARD);

  // Compute auto-correlation using FFT'd signal
  computeAutoCorrelation((complex*)pFFTBuf, g_tempBuffer3, g_bufferSize/2);

  // Calculate magnitudes
  for (int i=0; i< g_bufferSize; i+=2) {
    // main FFT buffer magnitude
    complex freqBin = *((complex*)(void*)&pFFTBuf[i]);
    pFFTBuf[i/2] = cmp_abs(freqBin);

    // auto-correlation signal magnitude
    freqBin = *((complex*)(void*)&g_tempBuffer3[i]);
    g_tempBuffer3[i/2] = cmp_abs(freqBin);
  }

  
  // Estimate pitch: by finding the bin with the highest
  // auto-correlation value 
  int maxBin = findMaxIndex(g_tempBuffer3, g_bufferSize/2);
  float maxMagnitude = g_tempBuffer3[maxBin];
  
  // Ignore pitch if the magnitude is low. Helps in ignoring
  // noise.
  if (maxMagnitude > PITCH_MAG_THRESH) {
    g_pitch= round((float)maxBin*g_sampleRate/g_bufferSize);
  } else {
    g_pitch = 0.0;
  }

  // Ignore very low frequencies - most likely DC or noise.
  if (g_pitch < MIN_PITCH) {
    g_pitch = 0;
  }


  // clear the color and depth buffers
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  // line width
  glLineWidth( 2.0 );

  // push the matrix
  glPushMatrix();

  // vertex coordinates
  GLfloat y;
  GLfloat x = X_START;
  GLfloat z = 1.0, zinc=-.1;

  // Calculate x increment
  GLfloat xinc;
  GLfloat logMax = log10(10.0+(GLfloat)g_bufferSize/2);
  if (!g_logFreq) {
    xinc = ::fabs(2*x / (g_bufferSize/2));
  }    

  // Select spectrum from history
  SAMPLE* pCurrentBuf = g_fftBuffer[g_BufIndex];
  int historyIndex = g_BufIndex;

  float pitchXLoc = 0.0;
  GLfloat colorDecay = 0.8;


  glColor3f(WATERFALL_COLOR(1.4));

  // Iterate over all spectrums in the history
  for( int i=0; i<FFT_HISTORY_LEN; i++) {
    pCurrentBuf = g_fftBuffer[historyIndex];
    GLfloat xFrac = 0.0, nextXFrac = 0.0;
    x = X_START;

  // Logarithmic x (frequency) axis  
  if (g_logFreq) {
      xFrac = getXFraction(0, logMax);
      nextXFrac = getXFraction(1, logMax);
      xinc = 2*(nextXFrac - xFrac);
    }

    // Start the line
    glBegin( GL_LINE_STRIP );        


    // loop through the spectrum
    for( int j = 0; j < g_bufferSize/2; j++ )
    {
      y = pCurrentBuf[j];
      y+= 10;
      y = log10(y);
      y = g_scaling_factor*(y-1.0);

      if (i==0 && j==maxBin) {
        pitchXLoc = x;
      }
      // Set the next vertex
      glVertex3f( x, y+1.5, z);

      // Logarithmic x (frequency) axis
      if (g_logFreq) {        
        xinc = ::fabs(3.159*X_START*(nextXFrac - xFrac));
        xFrac = nextXFrac;
        nextXFrac = getXFraction(j+1, logMax);
      }

      // Increment x
      x += xinc;
    }

    // Finish line
    glEnd();

    z += zinc;

    historyIndex--;
    if (historyIndex < 0) {
      historyIndex = FFT_HISTORY_LEN - 1;
    }

    // Fade out the color for older spectrums
    colorDecay *= COLOR_DECAY_RATE;
 
    // set the color for the next spectrum
    glColor3f(WATERFALL_COLOR(colorDecay));
  }

  // Draw a sphere at the fundemental pitch
  if (g_pitch) {
    // Push the current matrix
    glPushMatrix();

    // Color
    glColor3f( .5, 1.0, .5 );

    // Translate based on detected pitch
    glTranslatef( pitchXLoc, 1.45, 1 );

    // Sphere radium depends on the magnitude at of the bin
    // corresponding with the pitch
    float radius = maxMagnitude;
    radius += log10(radius+10);
    radius = (radius-1)*g_scaling_factor;
    radius = min( radius * 0.3, 0.1);

    // Draw sphere
    glutSolidSphere( radius, 20, 2 );

    // Pop matrix
    glPopMatrix();
  }

  // Next time we'll overwrite the oldest spectrum in the history
  g_BufIndex = (g_BufIndex + 1) % FFT_HISTORY_LEN;


  // ---- Draw windowed time domain signal ----------------

  // Calculate X increment
  xinc = ::fabs(2*x / g_bufferSize);

  // Start the line
  glBegin( GL_LINE_STRIP );


  glColor3f(WINDOWED_COLOR(1.0));
  x = X_START;

  // Iterate over samples
  for( int j = 0; j < g_bufferSize; j++ )
  {
    y = g_WindowedBuffer[j];

    // set the next vertex
    glVertex3f( x, y*5, 0);
    // increment x
    x += xinc;
  }

  // Finish line
  glEnd();
  // ------------------------------------------------------

  // pop
  glPopMatrix();

  // Flush
  glFlush();
  // Swap the double buffer
  glutSwapBuffers( );
}
